#include "gev_camera.hpp"
#include <QDebug>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc.hpp>
#include <iostream>
#include <iomanip>
#include <cstring>
#include <fstream>
#include <string>
#include <QString>

GEVCamera::GEVCamera(std::string macAddress)
    : macAddress(macAddress.c_str())
    , running(true)
    , system()
    , stream(NULL)
    , device(NULL)
    , bufferList()
    , deviceInfo()
    , imgWidth()
    , imgHeight()
    , imgBufferPtr(NULL)
    , threadPtr()
    , imgReadyConditionVariable()
    , imgReadyMutex()
    , mAcquisitionStateManager(NULL)
    , aEventSink(NULL)
{
}


/**************************************/
/*  connect()*/
/**************************************/

void GEVCamera::connect()
{

    if(this->threadPtr != nullptr)
    {
        std::cerr << "GEVCamera::connect: error: there is a connection running." << std::endl;
        return;
    }
    PvResult result = this->system.FindDevice(this->macAddress, &this->deviceInfo);
    this->device = PvDevice::CreateAndConnect(this->deviceInfo, &result);
    std::cout <<  "resum=" << result <<" "  <<  this->deviceInfo <<" "   <<  this->deviceInfo->GetDisplayID().GetAscii() << "\n" ;
    const PvInterface *pinterface; //
    pinterface=this->deviceInfo->GetInterface();

    QString Nom=this->deviceInfo->GetModelName().GetAscii();
    std::cout << Nom.toLocal8Bit().constData() << std::endl;
    std::cout <<  "nom de l'interface : " <<    pinterface->GetName().GetAscii() << std::endl;
    std::cout <<  "type de l'interface : " <<   pinterface->GetType() << std::endl;
    this->stream = this->openStream(deviceInfo);
    this->configureStream(device, stream);
    this->createStreamBuffers(this->device, stream, &this->bufferList);

    PvGenParameterArray *lDeviceParams = this->device->GetParameters();
    std::cout << "------------" << std::endl;

    std::cout << lDeviceParams << "------------ "<<lDeviceParams->SetEnumValue("PixelFormat",0x1100005) << std::endl; // OK Mono12  OK ça fonctionne
    NBITS=12;

    std::cout << "------------" << std::endl;
    PvGenParameter *lGenParameter4 = lDeviceParams->Get("MotorShutter_PulseRev"); //Ouverture shutter
    int64_t lValueR;
    static_cast<PvGenInteger *>( lGenParameter4 )->GetValue( lValueR );
    std::cout << "Integer: " << lValueR << std::endl;
    const PvString aValueR="200";
    PvResult ResultR=lGenParameter4->FromString(aValueR);	 //Ouverture shutter    //On remplit le registre avec la valeur
    std::cout << "Ouverture Shutter : " << ResultR	<< std::endl;

    // IL FAUT METTRE le pixeltype a Mono12bits
    PvGenParameter *lGenParameterP = lDeviceParams->Get("PixelFormat"); //PixelType
    int64 lValueRP;
    static_cast<PvGenEnum *>( lGenParameterP )->GetValue( lValueRP );
    std::cout << "PixelFormat: " << lValueRP << std::endl;

    PvGenEnum *lGenEnum=lDeviceParams->GetEnum("PixelFormat");
    PvString mastr=lGenEnum->ToString();

    std::cout << "GenEnum: " << mastr.GetAscii() << std::endl;
    lDeviceParams->GetEnumValue("PixelFormat",lValueRP);
    std::cout << "PixelFormat: " << lValueRP << std::endl;

    int64_t lValueRPO;
    lDeviceParams->GetIntegerValue("TestPattern",lValueRPO);
    PvGenEnum *lGenEnumT=lDeviceParams->GetEnum("TestPattern");

    PvResult ResultRT=lDeviceParams->SetEnumValue("TestPattern",0x0);  // Normal
    //PvResult ResultRT=lDeviceParams->SetEnumValue("TestPattern",0x1); //traits verticaux
    //PvResult ResultRT=lDeviceParams->SetEnumValue("TestPattern",0x2); // rampe

    std::cout << "TestPattern : " << ResultRT	<< std::endl;
    PvString mastrT=lGenEnumT->ToString();
    std::cout << "GenEnumT: " << mastrT.GetAscii() << std::endl;

    //------------------------
    // VITESSE ET TEMPS D'ACQUISITION !!!
    //------------------------
    // AcquisitionFrameTime  in us (= 1/(FPS) )
    // AcquisitionFrameRate (FPS) =10
    // ExposureTime (tint in us) =60000us=60ms
    double lValueFPS;
    lDeviceParams->GetFloatValue("AcquisitionFrameRate",lValueFPS);
    std::cout << "FPS " << lValueFPS << std::endl;
    lValueFPS=40;// // 10 FPS // 40 FPS max
    PvResult ResultFPS=lDeviceParams->SetFloatValue("AcquisitionFrameRate",lValueFPS);
    double lValuetint;
    lDeviceParams->GetFloatValue("ExposureTime",lValuetint);
    std::cout << ResultFPS <<" FPS " << lValueFPS << std::endl;
    std::cout << "tint " << lValuetint << std::endl;
    lValuetint=19000.0;
    PvResult Resulttint=lDeviceParams->SetFloatValue("ExposureTime",lValuetint);
    lDeviceParams->GetFloatValue("ExposureTime",lValuetint);
    std::cout << Resulttint <<"  tint " << lValuetint << std::endl;

    if (strcmp((Nom.toLocal8Bit().constData()),"FX10e")==0){
        std::cout << " ICI Fx10"<< std::endl;
    }
    else{
        std::cout << " ICI Fx17"<< std::endl;
    }


    PvGenParameterArray *lStreamParams = this->stream->GetParameters();

    PvGenParameter *lDeviceParamsRR = lStreamParams->Get("ResetOnIdle"); //

    int64 lvaleur=0;

    lStreamParams->GetIntegerValue("ResetOnIdle",lvaleur);
    std::cout << "ResetOnIdle " << lvaleur << std::endl;
    lvaleur=200;
    lStreamParams->SetIntegerValue("ResetOnIdle",lvaleur);

    lStreamParams->GetIntegerValue("ResetOnIdle",lvaleur);
    std::cout << "après ResetOnIdle " << lvaleur << std::endl;


    int64 lValueResetOnIdle=0;
    PvResult ResultResetonIdle=lStreamParams->SetIntegerValue("ResetOnIdle",lValueResetOnIdle);
    std::cout << ResultResetonIdle <<"   lValueResetOnIdle" << lValueResetOnIdle << std::endl;

    lStreamParams->GetIntegerValue("RequestTimeout",lvaleur);
    std::cout << " RequestTimeout" << lvaleur << std::endl;
    lvaleur=100;
    lStreamParams->SetIntegerValue("RequestTimeout",lvaleur);

    this->running = true;
    this->threadPtr = std::make_shared<std::thread>(&GEVCamera::acquireImages, this, this->device, this->stream);

} // FIN de connect()




/**************************************/
/*  openStream*/
/**************************************/



PvStream* GEVCamera::openStream(const PvDeviceInfo *aDeviceInfo)
{
    PvStream *lStream;
    PvResult lResult;

    // Open stream to the GigE Vision or USB3 Vision device
    std::cout << "Opening stream to device." << std::endl;
    lStream = PvStream::CreateAndOpen( aDeviceInfo->GetConnectionID(), &lResult );
    if ( lStream == NULL )
    {
        std::cout << "Unable to stream from " << aDeviceInfo->GetDisplayID().GetAscii() << "." << std::endl;
    }
    return lStream;
}



/**************************************/
/*  configureStream */
/**************************************/


void GEVCamera::configureStream(PvDevice* aDevice, PvStream* aStream)
{
    // If this is a GigE Vision device, configure GigE Vision specific streaming parameters
    PvDeviceGEV* lDeviceGEV = dynamic_cast<PvDeviceGEV *>( aDevice );
    if ( lDeviceGEV != NULL )
    {
        PvStreamGEV *lStreamGEV = static_cast<PvStreamGEV *>( aStream );
        // Negotiate packet size
        lDeviceGEV->NegotiatePacketSize();
        // Configure device streaming destination
        lDeviceGEV->SetStreamDestination( lStreamGEV->GetLocalIPAddress(), lStreamGEV->GetLocalPort() );
    }
}



/**************************************/
/*  createStreamBuffers*/
/**************************************/
void GEVCamera::createStreamBuffers(PvDevice* aDevice, PvStream* aStream, BufferList* aBufferList)
{
    // Reading payload size from device
    uint32_t lSize = aDevice->GetPayloadSize();

    // Use BUFFER_COUNT or the maximum number of buffers, whichever is smaller
    uint32_t lBufferCount = ( aStream->GetQueuedBufferMaximum() < BUFFER_COUNT ) ?
                aStream->GetQueuedBufferMaximum() :
                BUFFER_COUNT;

    std::cout << "lBufferCount: " << lBufferCount <<   " lSize=" << lSize  << std::endl;
    // Allocate buffers
    for ( uint32_t i = 0; i < lBufferCount; i++ )
    {
        // Create new buffer object
        PvBuffer *lBuffer = new PvBuffer;
        // Have the new buffer object allocate payload memory
        // lBuffer->Alloc( static_cast<uint32_t>( lSize ) );
        lBuffer->Alloc( static_cast<uint32_t>( lSize ) );
        //lImage->Alloc(lWidth, lHeight, PvPixelRGB8);
        // Add to external list - used to eventually release the buffers
        aBufferList->push_back( lBuffer );
    }

    // Queue all buffers in the stream
    BufferList::iterator lIt = aBufferList->begin();
    while ( lIt != aBufferList->end() )
    {
        aStream->QueueBuffer( *lIt );
        lIt++;
    }
} // FIN de createStreamBuffers


/**************************************/
/*  getImage()*/
/**************************************/

cv::Mat1w GEVCamera::getImage()
{
    std::unique_lock<std::mutex> lock(this->imgReadyMutex);
    this->imgReadyConditionVariable.wait(lock);
    //https://docs.opencv.org/3.3.1/d3/d63/classcv_1_1Mat.html#a51615ebf17a64c968df0bf49b4de6a3a
    return cv::Mat1w(this->imgHeight, this->imgWidth,this->imgBufferPtr);
} // FIN de getImage()


/**************************************/
/*  acquireImages*/
/**************************************/

void GEVCamera::acquireImages(PvDevice *aDevice, PvStream *aStream)
{
    // Get device parameters need to control streaming
    PvGenParameterArray *lDeviceParams = aDevice->GetParameters();

    // Map the GenICam AcquisitionStart and AcquisitionStop commands
    PvGenCommand *lStart = dynamic_cast<PvGenCommand *>(lDeviceParams->Get("AcquisitionStart"));

    // Get stream parameters
    PvGenParameterArray *lStreamParams = aStream->GetParameters();

    // Map a few GenICam stream stats counters
    PvGenFloat *lFrameRate = dynamic_cast<PvGenFloat *>(lStreamParams->Get("AcquisitionRate"));
    PvGenFloat *lBandwidth = dynamic_cast<PvGenFloat *>(lStreamParams->Get("Bandwidth"));

    // Enable streaming and send the AcquisitionStart command
    std::cout << "Enabling streaming and sending AcquisitionStart command." << std::endl;
    aDevice->StreamEnable();
    lStart->Execute();

    double lFrameRateVal = 0.0;
    double lBandwidthVal = 0.0;

    while(this->running)
    {
        PvBuffer *lBuffer = NULL;
        PvResult lOperationResult;

        // Retrieve next buffer
        PvResult lResult = aStream->RetrieveBuffer( &lBuffer, &lOperationResult, 5000 ); //1000 =timeout
        if ( lResult.IsOK() )
        {
            if ( lOperationResult.IsOK() )
            {
                PvPayloadType lType;
                //
                // We now have a valid buffer. This is where you would typically process the buffer.
                // -----------------------------------------------------------------------------------------
                // ...

                lFrameRate->GetValue( lFrameRateVal );
                lBandwidth->GetValue( lBandwidthVal );
                //qDebug() << "lFrameRateVal=" <<  lFrameRateVal;
              //  qDebug() <<  "lBandwidth=" <<  lBandwidth;
                // If the buffer contains an image, display width and height.
                lType = lBuffer->GetPayloadType();

                if ( lType == PvPayloadTypeImage )
                {
                    // Get image specific buffer interface.
                    PvImage *lImage = lBuffer->GetImage();
                    this->imgWidth = lImage->GetWidth();
                    this->imgHeight = lImage->GetHeight();
                    uint32_t imgSize = this->imgHeight*this->imgWidth;
                    if(this->imgBufferPtr == NULL)
                    {
                        this->imgBufferPtr = (uint16_t*)malloc(imgSize*2); // OUI  *2 (pour 16bits)
                    }
                    unsigned short *mimgptr = reinterpret_cast<ushort*>(lImage->GetDataPointer());
                    std::memcpy(this->imgBufferPtr, mimgptr, imgSize*2); // OUI  *2 (pour 16bits)

                    // http://opencv-users.1802565.n2.nabble.com/resizing-12-bit-data-td7583303.html
                    // https://stackoverflow.com/questions/25662617/loading-12-bit-raw-greyscale-image-using-opencv

                    // https://stackoverflow.com/questions/16575386/array-to-opencv-matrix
                    // cv::Mat cv_data(nrows, ncols, CV_64FC1, data);
                    // cv::Mat Img_Source16Bit_Gray(this->imgHeight,this->imgWidth,CV_16UC1,mimgptr);

                    //   std::ofstream fout16("/home/lepage/Bureau/Data16bits.txt");//
                    //    fout16 << cv::format(Img_Source16Bit_Gray,cv::Formatter::FMT_DEFAULT); //
                    //    fout16.close();

                    this->imgReadyConditionVariable.notify_one();
                }
                else
                {
                    std::cout << " (buffer does not contain image)";
                }
            }
            // Re-queue the buffer in the stream object
            aStream->QueueBuffer( lBuffer );
        }
        else
        {
            std::cout << "ligne 342 erreur : " << lResult.GetCodeString().GetAscii() << std::endl;
        }
    }
    qDebug() <<  "lBandwidth=" <<  lBandwidth;
    // Tell the device to stop sending images.
} //Fin de acquireImages()



/**************************************/
/*  shutter_open() */
/**************************************/
void GEVCamera::shutter_open(){

    PvGenParameterArray *lDeviceParamsb = this->device->GetParameters();
    PvGenParameter *lGenParameter4b = lDeviceParamsb->Get("MotorShutter_PulseRev"); //Ouverture shutter
    int64_t lValueR;
    const PvString aValueR="200";
    PvResult ResultR=lGenParameter4b->FromString(aValueR);
    //std::cout << "Ouverture Shutter : " << ResultR	<< std::endl;
    //qDebug() << " FIN GEV::shutter_open";

}


/**************************************/
/*  blanc() */
/**************************************/

void GEVCamera::blanc(){
    PvGenParameterArray *lDeviceParamsb = this->device->GetParameters();
    PvGenParameter *lGenParameter4b = lDeviceParamsb->Get("MotorShutter_PulseRev"); //Ouverture shutter
    int64_t lValueR;
    static_cast<PvGenInteger *>( lGenParameter4b )->GetValue( lValueR );
    std::cout << "Integer: " << lValueR << std::endl;
    const PvString aValueR="200";
    PvResult ResultR=lGenParameter4b->FromString(aValueR);	 //Ouverture shutter    //On remplit le registre avec la valeur
    std::cout << "Ouverture Shutter : " << ResultR	<< std::endl;
    ResultR=lGenParameter4b->FromString(aValueR);	 //Ouverture shutter    //On remplit le registre avec la valeur
    qDebug() << " FIN GEV::blanc";
}





/**************************************/
/*  darkcurrent() */
/**************************************/


void GEVCamera::darkcurrent(){

    PvGenParameterArray *lDeviceParams = this->device->GetParameters();
    PvGenParameter *lGenParameter = lDeviceParams->Get( 2 );
    PvString lValue;
    static_cast<PvGenString *>( lGenParameter )->GetValue( lValue );
    std::cout << "String: " << lValue.GetAscii()<< std::endl;

    PvGenParameter *lGenParameter3 = lDeviceParams->Get("MotorShutter_PulseFwd");

    int64_t lValueM;
    static_cast<PvGenInteger *>( lGenParameter3 )->GetValue( lValueM );
    std::cout << "Integer: " << lValueM << std::endl;
    const PvString aValue="200";
    PvResult Result=lGenParameter3->FromString(aValue);	//Fermeture shutter
    std::cout << "Fermeture Shutter : " << Result	<< std::endl;

} // FIN darkcurrent";



/**************************************/
/*  startstream() */
/**************************************/

void GEVCamera::startstream()
{
    PvGenParameterArray *lDeviceParams = this->device->GetParameters();
    PvGenCommand *lStart = dynamic_cast<PvGenCommand *>(lDeviceParams->Get("AcquisitionStart"));
    lStart->Execute();
    this->device->StreamEnable();
}



/**************************************/
/*  stopstream() */
/**************************************/

void GEVCamera::stopstream()
{
    PvGenParameterArray *lDeviceParams = this->device->GetParameters();
    PvGenCommand *lStop = dynamic_cast<PvGenCommand *>(lDeviceParams->Get("AcquisitionStop"));
    lStop->Execute();
    this->device->StreamDisable();
}


/**************************************/
/*  disconnect() */
/**************************************/

void GEVCamera::disconnect()
{
    qDebug() << " Début disconnect() ";
    this->threadPtr.get()->detach();
    this->running = false;
    PvGenParameterArray *lDeviceParams = this->device->GetParameters();
    PvGenCommand *lStop = dynamic_cast<PvGenCommand *>(lDeviceParams->Get("AcquisitionStop"));
    std::cout << "Sending AcquisitionStop command to the device" << std::endl;
    lStop->Execute();

    // Disable streaming on the device
    std::cout << "Disable streaming on the controller." << std::endl;
    this->device->StreamDisable();

    stream->Close();
    PvStream::Free( stream );
    PvDevice::Free( device );  // OK
    this->device=NULL;
    qDebug() << " Fin disconnect() ";
}
