#ifndef GEV_CAMERA_HPP
#define GEV_CAMERA_HPP

#include <PvSystem.h>
#include <PvInterface.h>
#include <PvDevice.h>
#include <PvStream.h>
#include <PvBuffer.h>
#include <PvStreamGEV.h>
#include <PvDeviceGEV.h>
#include <PvPixelType.h>
#include <PvBufferWriter.h>
#include <PvBufferConverterRGBFilter.h>
#include <QDebug>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc.hpp>
#include <list>
#include <thread>
#include <memory>
#include <mutex>
#include <condition_variable>
#include <PvPipeline.h>
#include <PvAcquisitionStateManager.h>

typedef std::list<PvBuffer *> BufferList;
#define BUFFER_COUNT ( 16 )

class GEVCamera
{
public:
    GEVCamera(std::string macAddress);

    void connect();

    void disconnect();

    PvStream* openStream(const PvDeviceInfo *aDeviceInfo);
    void configureStream(PvDevice* aDevice, PvStream* aStream);
    void createStreamBuffers(PvDevice* aDevice, PvStream* aStream, BufferList* aBufferList);
    void acquireImages(PvDevice *aDevice, PvStream *aStream);
    // cv::Mat1b getImage();
    cv::Mat1w getImage();

    void darkcurrent();

    void blanc();
    void shutter_open();
    void stopstream();
    void startstream();
    bool running;

    std::mutex imgReadyMutex;
    std::condition_variable imgReadyConditionVariable;

    PvAcquisitionStateManager *mAcquisitionStateManager;
    PvAcquisitionStateEventSink *aEventSink;

    PvString macAddress;
    PvSystem system;
    PvStream* stream;
    PvDevice* device;
    BufferList bufferList;
    uint32_t imgWidth;
    uint32_t imgHeight;
    cv::Mat imgMat;
    uint8_t* imgBufferPtr8;
    uint16_t* imgBufferPtr;
    int NBITS=12;

    const PvDeviceInfo* deviceInfo;

    std::shared_ptr<std::thread> threadPtr;
};

#endif // GEV_CAMERA_HPP
